"""
This module provides the main classes of the FakeNOS library.
It provides the FakeNOS class for creating a fake network
operating system and the Nos class for creating a network
operating system object.
"""

from fakenos.core.fakenos import FakeNOS
from fakenos.core.nos import Nos

__all__ = ("FakeNOS", "Nos")



"""
This module handles is the base model for any
server implemented as a plugin. To see an example
look for fakenos/plugins/servers/ssh_server_paramiko.py
"""

# pylint: disable=no-name-in-module
from abc import ABC, abstractmethod
import sys
import socket
import threading
import logging

log = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class TCPServerBase(ABC):
    """
    This module provides the base class for a TCP Server.
    It provides the methods to start and stop the server.

    Note: We are looking to switch to socketserver as it is
    the standard library in python.
    """

    def __init__(self, address="localhost", port=6000, timeout=1):
        """
        Initialize the server with the address and port
        and the timeout for the socket.
        """
        self.address = address
        self.port = port
        self.timeout = timeout
        self._is_running = threading.Event()
        self._socket = None
        self.client_shell = None
        self._listen_thread = None
        self._connection_threads = []

    def start(self):
        """
        Start Server which distributes the connections.
        It handles the creation of the socket, binding to the address and port,
        and starting the listening thread.
        """
        if self._is_running.is_set():
            return

        self._is_running.set()

        self._bind_sockets()

        self._listen_thread = threading.Thread(target=self._listen)
        self._listen_thread.start()

    def _bind_sockets(self):
        """
        It binds the sockets to the corresponding IPs and Ports.
        In Linux and OSX it reuses the port if needed but
        not in Windows
        """
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

        if sys.platform in ["linux"]:
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, True)

        self._socket.settimeout(self.timeout)
        self._socket.bind((self.address, self.port))

    def stop(self):
        """
        It stops the server joining the threads
        and closing the corresponding sockets.
        """
        if not self._is_running.is_set():
            return

        self._is_running.clear()
        self._listen_thread.join()
        self._socket.close()

        for connection_thread in self._connection_threads:
            connection_thread.join()

    def _listen(self):
        """
        This function is constantly running if the server is running.
        It waits for a connection, and if a connection is made, it will
        call the connection function.
        """
        while self._is_running.is_set():
            try:
                self._socket.listen()
                client, _ = self._socket.accept()
                connection_thread = threading.Thread(target=self.connection_function, args=(client, self._is_running,))
                connection_thread.start()
                self._connection_threads.append(connection_thread)
            except socket.timeout:
                pass

    @abstractmethod
    def connection_function(self, client):
        """
        This abstract method it is called when a new connection
        is made. The implementation should handle all the
        latter connection.
        """



"""
Network Operating Systems (NOS). Base class to build NOS plugins instances to use with FakeNOS.
"""

import logging
from typing import Optional, List
import importlib.util
import os
import yaml

from fakenos.core.pydantic_models import ModelNosAttributes

log = logging.getLogger(__name__)

available_platforms: List[str] = [
    "alcatel_aos",
    "alcatel_sros",
    "allied_telesis_awplus",
    "arista_eos",
    "aruba_os",
    "avaya_ers",
    "avaya_vsp",
    "broadcom_icos",
    "brocade_netiron",
    "checkpoint_gaia",
    "cisco_asa",
    "cisco_ftd",
    "cisco_ios",
    "cisco_nxos",
    "cisco_s300",
    "cisco_xr",
    "dell_force10",
    "dlink_ds",
    "eltex",
    "ericsson_ipos",
    "extreme_exos",
    "hp_comware",
    "hp_procurve",
    "huawei_smartax",
    "huawei_vrp",
    "ipinfusion_ocnos",
    "juniper_junos",
    "linux",
    "ubiquiti_edgerouter",
    "ubiquiti_edgeswitch",
    "vyatta_vyos",
    "yamaha",
    "zyxel_os",
]


class Nos:
    """
    Base class to build NOS plugins instances to use with FakeNOS.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name: str = "FakeNOS",
        commands: dict = None,
        initial_prompt: str = "FakeNOS>",
        filename: Optional[str] = None,
        dict_args: Optional[dict] = None,
    ) -> None:
        """
        Method to instantiate Nos Instance

        :param name: NOS plugin name
        :param commands: dictionary of NOS commands
        :param initial_prompt: NOS initial prompt
        """
        self.name = name
        self.commands = commands or {}
        self.initial_prompt = initial_prompt
        if filename:
            self.from_file(filename)
        elif dict_args:
            self.from_dict(dict_args)

        self.validate()

    def validate(self) -> None:
        """
        Method to validate NOS attributes: commands, name,
        initial prompt - using Pydantic models,
        raises ValidationError on failure.
        """
        ModelNosAttributes(**self.__dict__)
        log.debug("%s NOS attributes validation succeeded", self.name)

    def from_dict(self, data: dict) -> None:
        """
        Method to build NOS from dictionary data.

        Sample NOS dictionary::

            nos_plugin_dict = {
                "name": "MyFakeNOSPlugin",
                "initial_prompt": "{base_prompt}>",
                "commands": {
                    "terminal width 511": {
                        "output": "",
                        "help": "Set terminal width to 511",
                        "prompt": "{base_prompt}>",
                    },
                    "terminal length 0": {
                        "output": "",
                        "help": "Set terminal length to 0",
                        "prompt": "{base_prompt}>",
                    },
                    "show clock": {
                        "output": "MyFakeNOSPlugin system time is 00:00:00",
                        "help": "Show system time",
                        "prompt": "{base_prompt}>",
                    },
                },
            }

        :param data: NOS dictionary
        """
        self.name = data.get("name", self.name)
        self.commands.update(data.get("commands", self.commands))
        self.initial_prompt = data.get("initial_prompt", self.initial_prompt)

    def from_yaml(self, data: str) -> None:
        """
        Method to build NOS from YAML data.

        Sample NOS YAML file content::

            name: "MyFakeNOSPlugin"
            initial_prompt: "{base_prompt}>"
            commands:
                terminal width 511: {
                    "output": "",
                    "help": "Set terminal width to 511",
                    "prompt": "{base_prompt}>",
                }
                terminal length 0: {
                    "output": "",
                    "help": "Set terminal length to 0",
                    "prompt": "{base_prompt}>",
                }
                show clock: {
                    "output": "MyFakeNOSPlugin system time is 00:00:00",
                    "help": "Show system time",
                    "prompt": "{base_prompt}>",
                }

        :param data: YAML structured text
        """
        with open(data, "r", encoding="utf-8") as f:
            self.from_dict(yaml.safe_load(f))

        
    def from_module(self, data: str) -> None:
        """
        Method to import NOS data from python file or python module.

        Loads from the .py file using the recipe:
        https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly

        Sample Python NOS plugin file::

            name = "MyFakeNOSPlugin"

            INITIAL_PROMPT = "{base_prompt}>"

            commands = {
                "terminal width 511": {
                    "output": "",
                    "help": "Set terminal width to 511",
                    "prompt": "{base_prompt}>",
                },
                "terminal length 0": {
                    "output": "",
                    "help": "Set terminal length to 0",
                    "prompt": "{base_prompt}>",
                },
                "show clock": {
                    "output": "MyFakeNOSPlugin system time is 00:00:00",
                    "help": "Show system time",
                    "prompt": "{base_prompt}>",
                },
            }

        :param data: OS path string to Python .py file
        """
        # check if data is a .py file path
        if isinstance(data, str) and data.endswith(".py"):
            if os.path.isfile(data):
                spec = importlib.util.spec_from_file_location("nos_module", data)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
            else:
                raise FileNotFoundError(data)
        # blindly assume it is imported python module
        else:
            module = data
        # source attributes from loaded .py module
        self.name = getattr(module, "name", self.name)
        self.commands.update(getattr(module, "commands", self.commands))
        self.initial_prompt = getattr(module, "INITIAL_PROMPT", self.initial_prompt)
        
    def from_file(self, data: str) -> None:
        """
        Method to load NOS from YAML or Python file

        :param data: OS path string to `.yaml/.yml` or `.py` file with NOS data
        """
        if not self.is_file_ending_correct(data):
            raise ValueError(
                f'Unsupported "{data}" file extension.\
                              Supported: .py, .yml, .yaml'
            )
        if data.endswith((".yaml", ".yml")):
            self.from_yaml(data)
        elif data.endswith(".py"):
            self.from_module(data)

    def is_file_ending_correct(self, data: str) -> None:
        """
        Method to check if file extension is correct and load NOS data.
        Correct types are: .yaml, .yml and .py
        """
        return data.endswith((".yaml", ".yml", ".py"))



"""
NOS module for Cisco IOS
"""

import time

name = "cisco_ios"
INITIAL_PROMPT = "{base_prompt}>"


# pylint: disable=unused-argument
def make_show_clock(base_prompt, current_prompt, command):
    "Return String in format '*11:54:03.018 UTC Sat Apr 16 2022'"
    return time.strftime("*%H:%M:%S.000 %Z %a %b %d %Y")


RUNNING_CONFIGURATION = """
service timestamps debug datetime msec
service timestamps log datetime msec
no service password-encryption
!
hostname {base_prompt}
!
boot-start-marker
boot-end-marker
!
!
vrf definition MGMT
 rd 1:1
 !
 address-family ipv4
 exit-address-family
!
vrf definition abc-123
 rd 123:123
 !
 address-family ipv4
 exit-address-family
!
vrf definition def-456
 rd 456:456
 !
 address-family ipv4
 exit-address-family
!
!
aaa new-model
!
!
aaa authentication login default local
aaa authorization exec default local
!
!
aaa session-id common
!
!
clock timezone EET 2 0
mmi polling-interval 60
no mmi auto-configure
no mmi pvc
mmi snmp-timeout 180
!
!
no ip domain lookup
ip domain name lab.local
ip cef
no ipv6 cef
!
multilink bundle-name authenticated
!
!
archive
 path unix:BACKUP-
 write-memory
username cisco privilege 15 password 0 cisco
username nornir privilege 15 secret 5 $1$clqS$3CzLxkeHVJ9OYW83DOjuN/
!
redundancy
!
lldp run
!
!
!
interface Loopback0
 description Routing Loopback
 ip address 10.0.0.10 255.255.255.255
 ip ospf 1 area 0
 ipv6 address 2001::10/128
!
interface Loopback100
 ip address 1.1.1.100 255.255.255.255
!
interface Ethernet0/0
 description Main Interface for L3 features testing
 no ip address
 duplex auto
!
interface Ethernet0/0.102
 description to_vIOS1_Gi0/0.102
 encapsulation dot1Q 102
 ip address 10.1.102.10 255.255.255.0
 ipv6 address 2001:102::10/64
!
interface Ethernet0/0.107
 description to_IOL2_Eth0/0.107
 encapsulation dot1Q 107
 ip address 10.1.107.10 255.255.255.0
 ip ospf network point-to-point
 ip ospf 1 area 0
 ipv6 address 2001:107::10/64
!
interface Ethernet0/0.2000
 encapsulation dot1Q 2000
 vrf forwarding MGMT
 ip address 192.168.217.10 255.255.255.0
!
interface Ethernet0/1
 no ip address
 duplex auto
!
interface Ethernet0/2
 no ip address
 duplex auto
!
interface Ethernet0/3
 no ip address
 shutdown
 duplex auto
!
router ospf 1
 router-id 10.0.0.10
!
router bgp 100
 template peer-policy RR
  route-map rpl_in in
  route-map rpl_out out
  route-reflector-client
  next-hop-self
  maximum-prefix 1000 80
  send-community both
 exit-peer-policy
 !
 template peer-session RR
  remote-as 100
  update-source Loopback0
 exit-peer-session
 !
 bgp log-neighbor-changes
 neighbor peer_grpoup_a peer-group
 neighbor peer_grpoup_a remote-as 118
 neighbor peer_grpoup_a update-source Loopback0
 neighbor 10.0.0.4 inherit peer-session RR
 neighbor 10.0.0.8 inherit peer-session RR
 neighbor 10.1.48.4 peer-group peer_grpoup_a
 neighbor 10.1.48.4 description some peer description
 !
 address-family ipv4
  neighbor peer_grpoup_a send-community both
  neighbor peer_grpoup_a next-hop-self
  neighbor peer_grpoup_a route-map route-map-in in
  neighbor peer_grpoup_a route-map route-map-out out
  neighbor 10.0.0.4 activate
  neighbor 10.0.0.4 send-community extended
  neighbor 10.0.0.4 inherit peer-policy RR
  neighbor 10.0.0.8 activate
  neighbor 10.0.0.8 send-community extended
  neighbor 10.0.0.8 inherit peer-policy RR
  neighbor 10.1.48.4 activate
 exit-address-family
 !
 address-family vpnv4
  neighbor peer_grpoup_a send-community both
  neighbor peer_grpoup_a maximum-prefix 1000 80
  neighbor 10.0.0.4 activate
  neighbor 10.0.0.4 send-community extended
  neighbor 10.0.0.4 inherit peer-policy RR
  neighbor 10.0.0.8 activate
  neighbor 10.0.0.8 send-community extended
  neighbor 10.0.0.8 inherit peer-policy RR
  neighbor 10.1.48.4 activate
 exit-address-family
 !
 address-family ipv4 vrf abc-123
  neighbor 123.1.1.1 remote-as 321
  neighbor 123.1.1.1 activate
  neighbor 123.1.2.2 remote-as 321
  neighbor 123.1.2.2 activate
 exit-address-family
 !
 address-family ipv4 vrf def-456
  neighbor 251.1.1.5 remote-as 456
  neighbor 251.1.1.5 activate
  neighbor 251.6.1.6 remote-as 456
  neighbor 251.6.1.6 activate
 exit-address-family
!
ip forward-protocol nd
!
!
no ip http server
no ip http secure-server
ip route vrf MGMT 0.0.0.0 0.0.0.0 192.168.217.1 name inband_management
ip scp server enable
!
logging host 1.1.1.1
logging host 4.4.4.4
logging host 1.1.1.2
logging host 2.2.2.2
ipv6 ioam timestamp
!
control-plane
!

line con 0
 logging synchronous
line aux 0
line vty 0 4
 transport input telnet ssh
!
ntp server 8.8.8.8
ntp server 7.7.7.8
ntp server 1.1.1.2
ntp server 3.3.3.3
ntp server 7.7.7.7
!
end
"""

SHOW_VERSION = """
Cisco IOS XE Software, Version 17.03.01a
Cisco IOS Software [Amsterdam], Virtual XE Software \
    (X86_64_LINUX_IOSD-UNIVERSALK9-M), Version 17.3.1a, RELEASE SOFTWARE (fc3)
Technical Support: http://www.cisco.com/techsupport
Copyright (c) 1986-2020 by Cisco Systems, Inc.
Compiled Wed 12-Aug-20 00:16 by mcpre


Cisco IOS-XE software, Copyright (c) 2005-2020 by cisco Systems, Inc.
All rights reserved.  Certain components of Cisco IOS-XE software are
licensed under the GNU General Public License ("GPL") Version 2.0.  The
software code licensed under GPL Version 2.0 is free software that comes
with ABSOLUTELY NO WARRANTY.  You can redistribute and/or modify such
GPL code under the terms of GPL Version 2.0.  For more details, see the
documentation or "License Notice" file accompanying the IOS-XE software,
or the applicable URL provided on the flyer accompanying the IOS-XE
software.


ROM: IOS-XE ROMMON
{base_prompt} uptime is 1 day, 17 hours, 32 minutes
Uptime for this control processor is 1 day, 17 hours, 33 minutes
System returned to ROM by reload
System image file is "bootflash:packages.conf"
Last reload reason: reload



This product contains cryptographic features and is subject to United
States and local country laws governing import, export, transfer and
use. Delivery of Cisco cryptographic products does not imply
third-party authority to import, export, distribute or use encryption.
Importers, exporters, distributors and users are responsible for
compliance with U.S. and local country laws. By using this product you
agree to comply with applicable laws and regulations. If you are unable
to comply with U.S. and local laws, return this product immediately.

A summary of U.S. laws governing Cisco cryptographic products may be found at:
http://www.cisco.com/wwl/export/crypto/tool/stqrg.html

If you require further assistance please contact us by sending email to
export@cisco.com.

License Level: ax
License Type: N/A(Smart License Enabled)
Next reload license Level: ax

The current throughput level is 1000 kbps


Smart Licensing Status: UNREGISTERED/No Licenses in Use

cisco CSR1000V (VXE) processor (revision VXE) \
    with 715705K/3075K bytes of memory.
Processor board ID 9ESGOBARV9D
Router operating mode: Autonomous
3 Gigabit Ethernet interfaces
32768K bytes of non-volatile configuration memory.
3978420K bytes of physical memory.
6188032K bytes of virtual hard disk at bootflash:.

Configuration register is 0x2102
"""

commands = {
    "enable": {
        "output": None,
        "new_prompt": "{base_prompt}#",
        "help": "enter exec prompt",
        "prompt": INITIAL_PROMPT,
    },
    "show clock": {
        "output": make_show_clock,
        "help": "Display the system clock",
        "prompt": [INITIAL_PROMPT, "{base_prompt}#"],
    },
    "show running-config": {
        "output": RUNNING_CONFIGURATION,
        "help": "Current operating configuration",
        "prompt": "{base_prompt}#",
    },
    "show version": {
        "output": SHOW_VERSION,
        "help": "System hardware and software status",
        "prompt": "{base_prompt}#",
    },
    "_default_": {
        "output": "% Invalid input detected at '^' marker.",
        "help": "Output to print for unknown commands",
        "prompt": [INITIAL_PROMPT, "{base_prompt}#"],
    },
    "terminal width 511": {
        "output": "",
        "help": "Set terminal width to 511",
        "prompt": [INITIAL_PROMPT, "{base_prompt}#"],
    },
    "terminal length 0": {
        "output": "",
        "help": "Set terminal length to 0",
        "prompt": [INITIAL_PROMPT, "{base_prompt}#"],
    },
}



"""
FakeNOS Command Line Tool for running fake servers.
"""

import argparse
import logging

from fakenos import FakeNOS

__version__ = "1.0.0"

log = logging.getLogger(__name__)

DESCRIPTION_TEXT = """-i --inventory   OS Path to inventory file
"""

argparser = argparse.ArgumentParser(
    description=f"FakeNOS, version {__version__}",
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
opts = argparser.add_argument_group(description=DESCRIPTION_TEXT)

opts.add_argument(
    "-i",
    "--inventory",
    action="store",
    dest="INVENTORY",
    default=None,
    type=str,
    help=argparse.SUPPRESS,
)

opts.add_argument(
    "-l",
    "--log-level",
    action="store",
    dest="LOG_LEVEL",
    default="INFO",
    type=str,
    help="Log level",
)


args = argparser.parse_args()

logging.basicConfig(level=args.LOG_LEVEL.upper())


def run_cli():
    """Function to start FakeNOS CLI"""
    print("Initiating FakeNOS")
    fakenet = FakeNOS(inventory=args.INVENTORY)
    log.info("Initiating FakeNOS")
    fakenet.start()

    try:
        while True:
            pass
    except KeyboardInterrupt:
        log.info("Shutting down FakeNOS")
        fakenet.stop()


if __name__ == "__main__":
    run_cli()



"""
This module is the point of entry for all NOS plugins.
It imports all the NOS plugins and loads them into a dictionary.

The NOS plugins are loaded first, using the .py modules and
then the .yaml files in the platforms directory for those which are left.

With .py modules we can have functionality, while using YAML is only
intended for quick development of new NOS plugins.
"""

import glob
import os

from fakenos.core.nos import Nos

from . import cisco_ios
from . import arista_eos

nos_plugins = {}
nos_plugins_modules = {"cisco_ios": cisco_ios, "arista_eos": arista_eos}

current_file_path = os.path.abspath(__file__)
current_directory = os.path.dirname(current_file_path)
platforms_directory = os.path.join(current_directory, "platforms")
yaml_files = glob.glob(os.path.join(platforms_directory, "*.yaml"))

# load NOS from YAML files
for file in yaml_files:
    nos_instance = Nos()
    nos_instance.from_file(file)
    nos_plugins[nos_instance.name] = nos_instance

# load NOS from python modules updating the NOS
for plugin_name, plugin_module in nos_plugins_modules.items():
    if plugin_name in nos_plugins:
        nos_instance = nos_plugins[plugin_name]
    else:
        nos_instance = Nos()
    nos_instance.from_module(plugin_module)
    nos_plugins[nos_instance.name] = nos_instance
    


"""
File to contain pydantic models for plugins input/output data validation
"""

from __future__ import annotations
import sys

from typing import Union, Optional, List, Dict, Callable

from pydantic import model_validator, BaseModel, StrictStr, StrictInt, IPvAnyAddress

if sys.version_info >= (3, 8):
    from typing import Literal  # works with >=py3.8
else:
    from typing_extensions import Literal  # works with <py3.8I

# ---------------------------------------------------------------------------------------
# NOS plugin commands model
# ---------------------------------------------------------------------------------------


class ModelNosCommand(BaseModel):
    """
    Pydantic model for NOS command attributes.
    """

    output: Optional[Union[StrictStr, None, Callable]] = None
    help: Optional[StrictStr] = None
    prompt: Optional[Union[StrictStr, List[StrictStr]]] = None
    new_prompt: Optional[StrictStr] = None
    alias: Optional[StrictStr] = None


class ModelNosAttributes(BaseModel):
    """
    Pydantic model for NOS attributes.
    """

    commands: Dict[StrictStr, ModelNosCommand]
    name: StrictStr
    initial_prompt: StrictStr


class ModelHost(BaseModel):
    """
    Pydantic model for Host Attributes
    """

    name: StrictStr
    username: StrictStr
    password: StrictStr
    port: StrictInt
    platform: Optional[StrictStr] = None


# ---------------------------------------------------------------------------------------
# FakeNOS inventory data model components
# ---------------------------------------------------------------------------------------


class NosPluginConfig(BaseModel):
    """
    Pydantic model for NOS plugin configuration.
    """

    commands: Optional[Dict[StrictStr, ModelNosCommand]] = None


class NosPlugin(BaseModel):
    """
    Pydantic model for NOS plugin.
    """

    plugin: StrictStr
    configuration: Optional[NosPluginConfig] = None


class ParamikoSshServerConfig(BaseModel):
    """
    Pydantic model for Paramiko SSH server configuration.
    """

    ssh_key_file: Optional[StrictStr] = None
    ssh_key_file_password: Optional[StrictStr] = None
    ssh_banner: Optional[StrictStr] = "FakeNOS Paramiko SSH Server"
    timeout: Optional[StrictInt] = 1
    address: Optional[Union[Literal["localhost"], IPvAnyAddress]] = None
    watchdog_interval: Optional[StrictInt] = 1


class ParamikoSshServerPlugin(BaseModel):
    """
    Pydantic model for Paramiko SSH server plugin.
    """

    plugin: Literal["ParamikoSshServer"]
    configuration: Optional[ParamikoSshServerConfig] = None


class CMDShellConfig(BaseModel):
    """
    Pydantic model for CMD shell configuration.
    """

    intro: Optional[StrictStr] = "Custom SSH Shell"
    ruler: Optional[StrictStr] = ""
    completekey: Optional[StrictStr] = "tab"
    newline: Optional[StrictStr] = "\r\n"


class CMDShellPlugin(BaseModel):
    """
    Pydantic model for CMD shell plugin.
    """

    plugin: Literal["CMDShell"]
    configuration: Optional[CMDShellConfig] = None


class InventoryDefaultSection(BaseModel):
    """
    Pydantic model for FakeNOS inventory default section.
    """

    username: Optional[StrictStr] = None
    password: Optional[StrictStr] = None
    # port: Optional[Union[conint(strict=True, gt=0, le=65535),
    # conlist(conint(strict=True, gt=0, le=65535),
    # min_items=2, max_items=2, unique_items=True)]]
    # use this for now, mkdocstring having issue with pydantic
    # https://github.com/mkdocstrings/griffe/issues/66
    port: Optional[Union[StrictInt, List[StrictInt]]] = None
    server: Optional[Union[ParamikoSshServerPlugin]] = None
    shell: Optional[Union[CMDShellPlugin]] = None
    nos: Optional[NosPlugin] = None


class HostConfig(InventoryDefaultSection):
    """
    Pydantic model for FakeNOS inventory host configuration.
    """

    # count: Optional[conint(strict=True, gt=0)]
    # use this for now, mkdocstring having issue with pydantic
    # https://github.com/mkdocstrings/griffe/issues/66
    replicas: Optional[StrictInt] = None

    @model_validator(mode="before")
    @classmethod
    def check_port_value(cls, values):
        """
        Method to validate port value based on 'replicas' value.
        """
        port = values.get("port")
        if "replicas" not in values and port:
            assert isinstance(port, int), "If no host 'replicas' given, port must be an integer"
        elif "replicas" in values and port:
            assert isinstance(port, list), "If host 'replicas' given, port must be a list"
        return values


class ModelFakenosInventory(BaseModel):
    """FakeNOS inventory data schema"""

    default: Optional[InventoryDefaultSection] = None
    hosts: Dict[StrictStr, HostConfig]

    # pylint: disable=too-few-public-methods
    class ConfigDict:
        """Pydantic model configuration"""

        extra = "forbid"



"""
This module implements an SSH server done using
paramiko as the SSH connection library.
"""

import logging
import io
import threading
import time
import socket

import paramiko
import paramiko.channel
import paramiko.transport

from fakenos.core.servers import TCPServerBase

log = logging.getLogger(__name__)

DEFAULT_SSH_KEY = """-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAnahBtR7uxtHmk5UwlFfpC/zxdxjUKPD8UpNOOtIJwpei7gaZ
+Jgub5GFJtTG6CK+DIZiR4tE9JxMjTEFDCGA3U4C36shHB15Pl3bLx+UxdyFylpc
c7XYp4fpQjhFUoHOAIl5ZaA223kIxi7sFXtM1Gjy6g49u+G5teVfMbeZnks2xjjy
F84qVADFBXCsfjrY5m4R+Wnfups/jP1agOpnOvqHlX/bpvzEZRcwJ0A8CylBZzQP
D1Y4EXy1B4QLyLJKFIMRkWnr0f8rK5Q/obCLTjl+IMmZrkItbfC/hYCy6TDi+Efn
cgGw02L93Mf6QGDNc21BsRELPYMME22MmpLphQIBIwKCAQEAmScbQjtOWr1GY3r7
/dG90SGaG+w70AALDmM2DUEQy6k/MF4vLAGMMd3RzfNE4YDV4EgHszbVRWSiIsHn
pWJf7OyyVZ7s9r2LuO111gFr82iB98V+YcaX8zOSIxIXdLicOwk0GZRSjA8tGErW
tcg8AYqFkulDSMylxqRN2IZ3+NnTROxh4uUFH57roSYoCvzjM2v1Xa+S42BLpBD1
3mLAJD36JhOhMTgYUgHAROx9+YUUUzYk3jpkTGWnAYSumnJXQYphLE9zadXxh94N
HZJdvXajuP5N2M3Q2b4Gbyt2wNFlNcHGA+Zwk8wHIBnY9Sb9Gz0QALsOAwUoRY8T
rCysSwKBgQDPVjFdSgM3jScmFV9fVnx3iNIlM6Ea7+UCrOOCvcGtzDo5vuTPktw7
8abHEFHw7VrtxI3lRQ41rlmK3B//Q7b+ZJ0HdZaRdyCqW1u91tq1tQe7yiJBm0c5
hZ3F0Vr6HAXoBVOux5wUq55jvUJ8dCVYNYfctZducVmOos3toDkSzQKBgQDCqRQ/
GO5AU3nKfuJ+SZvv8/gV1ki8pGmyxkSebUqZSXFx+rQEQ1e6tZvIz/rYftRkXAyL
XfzXX8mU1wEci6O1oSLiUBgnT82PtUxlO3Peg1W/cpKAaIFvvOIvUMRGFbzWhuj7
4p4KJjZWjYkAV2YlZZ8Br23DFFjjCuawX7NhmQKBgHCN4EiV5H09/08wLHWVWYK3
/Qzhg1fEDpsNZZAd3isluTVKXvRXCddl7NJ2kuHf74hjYvjNt0G2ax9+z4qSeUhF
P00xNHraRO7D4VhtUiggcemZnZFUSzx7vAxNFCFfq29TWVBAeU0MtRGSoG9yQCiS
Fo3BqfogRo9Cb8ojxzYXAoGBAIV7QRVS7IPheBXTWXsrKRmRWaiS8AxTe63JyKcm
XwoGea0+MkwQ67M6s/dqCxgcdGITO81Hw1HbSGYPxj91shYlWb/B5K0+CUyZk3id
y8vHxcUbXSTZ8ls/sQqAhpZ1Tkn2HBpvglAaM+OUQK/G5vUSe6liWeTawJuvtCEr
rjRLAoGAUNNY4/7vyYFX6HkX4O2yL/LZiEeR6reI9lrK/rSA0OCg9wvbIpq+0xPG
jCrc8nTlA0K0LtEnE+4g0an76nSWUNiP4kALROfZpXajRRaWdwFRAO17c9T7Uxc0
Eez9wYRqHiuvU0rryYvGyokr62w1MtJO0tttnxe1Of6wzb1WeCU=
-----END RSA PRIVATE KEY-----"""


class ParamikoSshServerInterface(paramiko.ServerInterface):
    """
    Class to implement the SSH server interface
    using paramiko as the SSH connection library.
    """

    def __init__(
        self,
        ssh_banner="FakeNOS Paramiko SSH Server",
        username=None,
        password=None,
    ):
        self.ssh_banner = ssh_banner
        self.username = username
        self.password = password

    def check_channel_request(self, kind, chanid):
        """
        This will allow the SSH server to provide a channel for the client
        to communicate over. By default, this will return
        OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED, so  we have to override it
        to return OPEN_SUCCEEDED when the kind of channel
        requested is "session".
        """
        if kind == "session":
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    # pylint: disable=too-many-arguments
    def check_channel_pty_request(self, channel, term, width, height, pixelwidth, pixelheight, modes):
        """
        AFAIK, pty (pseudo-tty (TeleTYpewriter))
        will allow our client to interact with our shell.
        """
        return True

    def check_channel_shell_request(self, channel):
        """
        This allows us to provide the channel
        with a shell we can connect to it.
        """
        return True

    def check_auth_password(self, username, password):
        if (username == self.username) and (password == self.password):
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def get_banner(self):
        """
        String that will display when a client connects,
        before authentication has happened. This is different
        than the shell's intro property, which is displayed
        after the authentication.
        """
        return (self.ssh_banner + "\r\n", "en-US")


class TapIO(io.StringIO):
    """
    Class to implement StringIO subclass but with blocking readline method
    and a list to buffer lines on write
    """

    def __init__(self, run_srv, initial_value="", newline="\n"):
        self.lines = []
        self.run_srv = run_srv
        super().__init__(initial_value, newline)

    def readline(self):
        """method to readline in indefinite block mode"""
        while self.run_srv.is_set():
            if self.lines:
                return self.lines.pop(-1)
            time.sleep(0.01)
        return None

    def write(self, value):
        """
        :param value: line to add to self.lines buffer
        """
        self.lines.insert(0, value)


def channel_to_shell_tap(channel_stdio, shell_stdin, shell_replied_event, run_srv):
    """
    Method to tap into the channel_stdio and send it to the shell
    """
    buffer = io.BytesIO()
    while run_srv.is_set():
        byte = channel_stdio.read(1)
        log.debug("ssh_server.channel_to_shell_tap received from channel: %s", [byte])
        if byte in (b"\r", b"\n"):
            shell_replied_event.wait(10)
            if not channel_stdio.channel.active:
                log.error("SSH channel is not active. Exiting.")
                break
            log.debug("ssh_server.channel_to_shell_tap echoing new line to channel: %s", [b"\r\n"])
            try:
                channel_stdio.write(b"\r\n")
            except (OSError, EOFError) as e:
                log.error("ssh_server.channel_to_shell_tap channel write error: %s", e)
                break
            buffer.write(byte)
            buffer.seek(0)
            line = buffer.read().decode(encoding="utf-8")
            buffer.seek(0)
            buffer.truncate()
            log.debug("ssh_server.channel_to_shell_tap sending line to shell: %s", [line])
            shell_stdin.write(line)
            shell_replied_event.clear()
        else:
            shell_replied_event.wait(10)
            if not channel_stdio.channel.active:
                log.error("SSH channel is not active. Exiting.")
                break
            try:
                channel_stdio.write(byte)
            except (OSError, EOFError) as e:
                log.error("ssh_server.channel_to_shell_tap channel write error: %s", e)
                break
            if byte not in [b"\x00", b""]:
                buffer.write(byte)
        time.sleep(0.01)


def shell_to_channel_tap(
    channel_stdio: paramiko.channel.ChannelFile,
    shell_stdout: TapIO,
    shell_replied_event: threading.Event,
    run_srv: threading.Event,
):
    """
    Method to tap into the shell_stdout and send it to the channel
    """
    while run_srv.is_set():
        if channel_stdio.closed:
            break
        line = shell_stdout.readline()
        # line is None we break the loop
        if line is None:
            break
        if "\r\n" not in line and "\n" in line:
            line = line.replace("\n", "\r\n")
        log.debug("ssh_server.shell_to_channel_tap sending line to channel %s", [line])
        try:
            channel_stdio.write(line.encode(encoding="utf-8"))
        except socket.error as e:
            if e.errno == 104:  # Connection reset by peer
                log.error("ssh_server.shell_to_channel_tap channel write error: %s", e)
                break
        shell_replied_event.set()


class ParamikoSshServer(TCPServerBase):
    """
    Class to implement an SSH server using paramiko
    as the SSH connection library.
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        shell,
        nos,
        nos_inventory_config,
        port,
        username,
        password,
        ssh_key_file=None,
        ssh_key_file_password=None,
        ssh_banner="FakeNOS Paramiko SSH Server",
        shell_configuration=None,
        address="127.0.0.1",
        timeout=1,
        watchdog_interval=1,
    ):
        super().__init__()

        self.nos = nos
        self.nos_inventory_config = nos_inventory_config
        self.shell = shell
        self.shell_configuration = shell_configuration or {}
        self.ssh_banner = ssh_banner
        self.username = username
        self.password = password
        self.port = port
        self.address = address
        self.timeout = timeout
        self.watchdog_interval = watchdog_interval

        if ssh_key_file:
            self._ssh_server_key = paramiko.RSAKey.from_private_key_file(ssh_key_file, ssh_key_file_password)
        else:
            self._ssh_server_key = paramiko.RSAKey(file_obj=io.StringIO(DEFAULT_SSH_KEY))

    def watchdog(self, is_running, run_srv, session, shell):
        """
        Method to monitor server liveness and recover where possible.
        """
        while run_srv.is_set():
            # check if session is alive, stop shell if it is not
            if not session.is_alive():
                log.warning(
                    "ParamikoSshServer.watchdog - \
                        session not alive, stopping shell"
                )
                shell.stop()
                break
            
            # exit the shell
            if not is_running.is_set():
                shell.stop()

            time.sleep(self.watchdog_interval)

    def connection_function(self, client, is_running):
        shell_replied_event = threading.Event()
        run_srv = threading.Event()
        run_srv.set()

        # create the SSH transport object
        session = paramiko.Transport(client)
        session.add_server_key(self._ssh_server_key)

        # create the server
        server = ParamikoSshServerInterface(
            ssh_banner=self.ssh_banner,
            username=self.username,
            password=self.password,
        )

        # start the SSH server
        session.start_server(server=server)

        # create the channel and get the stdio
        channel = session.accept()
        channel_stdio = channel.makefile("rw")

        # create stdio for the shell
        shell_stdin, shell_stdout = TapIO(run_srv), TapIO(run_srv)

        # start intermediate thread to tap into
        # the channel_stdio->shell_stdin bytes stream
        channel_to_shell_tapper = threading.Thread(
            target=channel_to_shell_tap,
            args=(channel_stdio, shell_stdin, shell_replied_event, run_srv),
        )
        channel_to_shell_tapper.start()

        # start intermediate thread to tap into
        # the shell_stdout->channel_stdio bytes stream
        shell_to_channel_tapper = threading.Thread(
            target=shell_to_channel_tap,
            args=(channel_stdio, shell_stdout, shell_replied_event, run_srv),
        )
        shell_to_channel_tapper.start()

        # create the client shell
        client_shell = self.shell(
            stdin=shell_stdin,
            stdout=shell_stdout,
            nos=self.nos,
            nos_inventory_config=self.nos_inventory_config,
            is_running=is_running,
            **self.shell_configuration,
        )

        # start watchdog thread
        watchdog_thread = threading.Thread(target=self.watchdog, args=(is_running, run_srv, session, client_shell))
        watchdog_thread.start()

        # running this command will block this function until shell exits
        client_shell.start()
        log.debug("ParamikoSshServer.connection_function stopped shell thread")

        # kill this server threads - watchdog, TapIO,
        # shell_to_channel_tapper and channel_to_shell_tapper
        run_srv.clear()
        log.debug("ParamikoSshServer.connection_function stopped server threads")

        # After execution continues, we can close the session
        session.close()
        log.debug("ParamikoSshServer.connection_function closed transport %s", session)



"""
Main module to interact with FakeNOS servers.
It is the entry point to start, stop and list FakeNOS servers.
"""

import logging
import copy
import threading
import time
import platform
from typing import Union, List, Dict, Set

import yaml
import detect

from fakenos.core.host import Host
from fakenos.core.nos import Nos
from fakenos.core.pydantic_models import ModelFakenosInventory

from fakenos.plugins.servers import servers_plugins
from fakenos.plugins.nos import nos_plugins
from fakenos.plugins.shell import shell_plugins


log = logging.getLogger(__name__)

default_inventory = {
    "default": {
        "username": "user",
        "password": "user",
        "port": 6000,
        "server": {
            "plugin": "ParamikoSshServer",
            "configuration": {
                "address": "127.0.0.1",
                "timeout": 1,
            },
        },
        "shell": {"plugin": "CMDShell", "configuration": {}},
        "nos": {"plugin": "cisco_ios", "configuration": {}},
    },
    "hosts": {
        "router0": {"port": 6000, "platform": "cisco_ios"},
        "router1": {"port": 6001, "platform": "arista_eos"}
    },
}

# If Windows or WSL, the configuration address is 0.0.0.0
# WSL Bug: https://github.com/microsoft/WSL/issues/4983
if detect.docker and "WSL2" in platform.release():
    server_config = default_inventory["default"]["server"]["configuration"]
    server_config["address"] = "0.0.0.0"


class FakeNOS:
    """
    FakeNOS class is a main entry point to interact
    with fake NOS servers - start, stop, list.

    :param inventory: FakeNOS inventory dictionary or
                      OS path to .yaml file with inventory data
    :param plugins: Plugins to add extra devices/commands
                    currently not supported easily.

    Sample usage:

    ```python
    from fakenos import FakeNOS

    net = FakeNOS()
    net.start()
    ```
    """

    def __init__(
        self,
        inventory: dict = None,
        plugins: list = None,
    ) -> None:
        self.inventory: dict = inventory or default_inventory
        self.plugins: list = plugins or []

        self.hosts: Dict[str, Host] = {}
        self.allocated_ports: Set[str] = set()

        self.shell_plugins = shell_plugins
        self.nos_plugins = nos_plugins
        self.servers_plugins = servers_plugins

        self._load_inventory()
        self._init()
        self._register_nos_plugins()

    def _is_inventory_in_yaml(self) -> bool:
        """method that checks if the inventory is a yaml file."""
        return isinstance(self.inventory, str) and self.inventory.endswith(".yaml")

    def _load_inventory_yaml(self) -> None:
        """Helper method to load FakeNOS inventory if it is yaml."""
        with open(self.inventory, "r", encoding="utf-8") as f:
            self.inventory = yaml.safe_load(f.read())

    def _load_inventory(self) -> None:
        """Helper method to load FakeNOS inventory"""
        if self._is_inventory_in_yaml():
            self._load_inventory_yaml()

        self.inventory["default"] = {
            **default_inventory["default"],
            **self.inventory.get("default", {}),
        }

        ModelFakenosInventory(**self.inventory)
        log.debug("FakeNOS inventory validation succeeded")

    def _init(self) -> None:
        """
        Helper method to initiate host objects
        and store them in self.hosts, this
        method called automatically on FakeNOS object instantiation.
        """
        for host_name, host_config in self.inventory["hosts"].items():
            params = {
                **copy.deepcopy(self.inventory["default"]),
                **copy.deepcopy(host_config),
            }
            port: Union[int, list] = params.pop("port")
            replicas: int = params.pop("replicas", None)
            self._check_ports_and_replicas_are_okey(port, replicas)
            self._instantiate_host_object(host_name, port, replicas, params)

    def _check_ports_and_replicas_are_okey(self, port, replicas):
        """
        Method to check if the port and replicas are okey

        :param port: integer or list of two integers - port to allocate
        :param replicas: integer - number of hosts to create
        """
        if not replicas and isinstance(port, list):
            raise ValueError("If replicas is not set, port must be an integer.")
        if replicas and not isinstance(port, list):
            raise ValueError("If replicas is set, port must be a list of two integers.")
        if replicas and len(port) != 2:
            raise ValueError("If replicas is set, port must be a list of two integers.")
        if replicas and port[0] >= port[1]:
            raise ValueError("If replicas is set, port[0] must be less than port[1].")
        if replicas and replicas < 1:
            raise ValueError("If replicas is set, replicas must be greater than 0.")
        if replicas and port[1] - port[0] + 1 != replicas:
            raise ValueError(
                "If replicas is set, port range \
                    must be equal to the number of replicas."
            )

    def _instantiate_host_object(self, host_name: str, port: Union[int, List[int]], replicas: int, params: dict):
        """
        Method that instantiate the host objects. It initializes the hosts
        with the corresponding name, port and network operating system

        :param host: string - name of the host
        :param port: integer or list of two integers - port to allocate
        :param count: integer - number of hosts to create
        :param params: dictionary - parameters to pass to
                                    the host like configurations
        """
        hosts_name, ports = self._get_hosts_and_ports(host_name, port, replicas)
        for h_name, p in zip(hosts_name, ports):
            self._instantiate_single_host_object(h_name, p, params)

    def _get_hosts_and_ports(self, host_name: str, port: Union[int, List[int]], replicas: int = None):
        """
        Method to get hosts and ports correctly
        depending on the number of replicas (if exists).

        :param host_name: string - name of the host
        :param port: integer or list of two integers - port to allocate
        :param replicas: integer - number of hosts to create
        """
        hosts_name: Set[str] = {}
        ports: Set[int] = {}

        if replicas:
            hosts_name = {f"{host_name}{i}" for i in range(replicas)}
            ports = set(range(port[0], port[1] + 1))
        else:
            hosts_name = {host_name}
            ports = {port}
        return hosts_name, ports

    def _instantiate_single_host_object(self, host, port, params):
        """
        Method that instantiate the host objects. It initializes the hosts

        :param host: string - name of the host
        :param port: integer or list of two integers - port to allocate
        :param params: dictionary - parameters to pass to
                                    the host like configurations
        """
        self._allocate_port(port)
        self.hosts[host] = Host(name=host, port=port, fakenos=self, **params)

    def _allocate_port(self, port: Union[int, List[int]]) -> None:
        """
        Method to allocate port for host

        :param port: integer or list of two integers -
                     range to allocate port from
        """
        if isinstance(port, int):
            port: List[int] = [port]

        for p in port:
            allocated_port = self._allocate_port_single(p)
            self.allocated_ports.add(allocated_port)

    def _allocate_port_single(self, port: int) -> int:
        """
        Method to allocate single port for host.

        :param port: integer - port to allocate
        """
        if port in self.allocated_ports:
            raise ValueError(f"Port {port} already in use")
        self.allocated_ports.add(port)
        return port

    def _get_hosts_as_list(self, hosts: Union[str, List[str]] = None) -> List[Host]:
        """
        Helper method to get hosts as list

        :param hosts: string or list of strings
        :return: list of strings
        """
        hosts_list: List[Host] = []
        if not hosts:
            hosts = list(self.hosts.keys())
        if isinstance(hosts, str):
            hosts = [hosts]
        hosts_list = [self.hosts[host] for host in hosts]
        return hosts_list

    def start(self, hosts: Union[str, list] = None) -> None:  # type: ignore
        """
        Function to start NOS servers instances

        :param hosts: single or list of hosts to start by their name.
        """
        hosts: List[str] = self._get_hosts_as_list(hosts)
        self._execute_function_over_hosts(hosts, "start", host_running=False)
        log.info("The following devices has been initiated: %s", [host.name for host in hosts])

    def stop(self, hosts: Union[str, List[str]] = None) -> None:
        """
        Function to stop NOS servers instances. It waits 2 seconds
        just in case that there is any thread doing something.

        :param hosts: single or list of hosts to stop by their name.
        """
        hosts: List[str] = self._get_hosts_as_list(hosts)
        self._execute_function_over_hosts(hosts, "stop", host_running=True)
        if hosts == list(self.hosts.values()):
            self._join_threads()

    def _join_threads(self) -> None:
        """
        Method to join threads in case that all hosts are stopped.
        """
        all_threads = threading.enumerate()
        for thread in all_threads:
            if thread is not threading.main_thread() and "pytest_timeout" not in thread.name:
                thread.join()
        n_threads: int = 2 if detect.windows else 1
        while threading.active_count() > n_threads:
            time.sleep(0.01)

    def _execute_function_over_hosts(self, hosts: List[Host], func: str, host_running: bool = True):
        """
        Function that executes a function like start or stop over
        the selected hosts.

        :param hosts: list of Hosts objects in which the function will
        be executed.
        """
        for host in hosts:
            if host not in self.hosts.values():
                raise ValueError(f"Host {host} not found")
            if host.running == host_running:
                getattr(host, func)()

    def _register_nos_plugins(self) -> None:
        """
        Method to register NOS plugin with FakeNOS object, all plugins
        must be registered before calling start method.

        :param plugin: OS path string to NOS plugin `.yaml/.yml` or `.py` file,
          dictionary or instance if Nos class
        """
        for plugin in self.plugins:
            if isinstance(plugin, Nos):
                nos_instance = plugin
            else:
                nos_instance = Nos()
                if isinstance(plugin, dict):
                    nos_instance.from_dict(plugin)
                elif isinstance(plugin, str):
                    nos_instance.from_file(plugin)
                else:
                    raise TypeError(f"Unsupported NOS type {type(plugin)}, supported str, dict or Nos")
            self.nos_plugins[nos_instance.name] = nos_instance



"""
NOS module for Arista EOS
"""

import time

name = "arista_eos"
INITIAL_PROMPT = "{base_prompt}>"


# pylint: disable=unused-argument
def make_show_clock(base_prompt, current_prompt, command):
    """Return the current time."""
    return f"""{time.strftime('%a %b %d %H:%M:%S %Y')}
Timezone: UTC
Clock source: local"""


# pylint: disable=unused-argument
def make_exit(base_prompt, current_prompt, command):
    """Exit the current level of the CLI."""
    if current_prompt in [f"{base_prompt}>", f"{base_prompt}#"]:
        return True  # close session
    if current_prompt in [f"{base_prompt}(config)#"]:
        return {"output": "", "new_prompt": "{base_prompt}#"}  # return to exec prompt
    raise RuntimeError(f"make_exit does not know how to handle '{current_prompt}' prompt")


RUNNING_CONFIGURATION = """! Command: show running-config
! device: {base_prompt} (cEOSLab, EOS-4.26.0F-21792469.4260F (engineering build))
!
transceiver qsfp default-mode 4x10G
!
service routing protocols model ribd
!
agent Bfd shutdown
agent PowerManager shutdown
agent LedPolicy shutdown
agent Thermostat shutdown
agent PowerFuse shutdown
agent StandbyCpld shutdown
agent LicenseManager shutdown
!
no logging console
logging host 1.1.1.1 514
logging host 1.2.3.4 514
logging host 1.2.3.5 514
logging host 1.2.3.6 514
logging host 2.2.2.2 514
logging host 4.3.2.1 514
logging host 5.5.5.5 514
logging host 7.7.7.7 514
logging host 9.9.9.9 514
!
logging level AAA informational
!
hostname {base_prompt}
!
ntp server 1.1.1.1
ntp server 1.1.1.2
ntp server 1.1.1.10
ntp server 1.1.1.11
ntp server 2.2.2.2
ntp server 2.2.2.3
ntp server 3.3.3.3
ntp server 3.3.3.4
ntp server 4.3.2.1
ntp server 6.6.6.6
ntp server 7.7.7.7
ntp server 7.7.7.8
!
snmp-server location "North West Hall DC1"
snmp-server local-interface Ethernet1
snmp-server host 1.2.3.4 version 2c test
snmp-server host 1.2.3.5 version 2c test
!
spanning-tree mode mstp
!
no aaa root
!
username nornir privilege 15 secret sha512 $6$EkriX8oB5g3Midq4$ErOpqzIWT7FxiW1IkSNQKS8gEqsn9HsbRVm8.Zw47y3Xm9a.GywP9zPF/avyTBBS8c5/ZSMMj/6BHL64KcW2I1
!
interface Ethernet1
   description Configured by NETCONF
   mtu 9200
   no switchport
   ip address 10.0.1.4/24
!
interface Loopback1
   ip address 1.1.1.1/24
!
interface Loopback2
   description Lopback2 for Customer 27123
   ip address 2.2.2.2/24
!
interface Loopback3
   description Customer #56924 service
   ip address 1.2.3.4/24
!
no ip routing
!
management api http-commands
   protocol http
   no shutdown
!
management api gnmi
   transport grpc default
!
management api netconf
   transport ssh def
!
end"""

SHOW_VERSION = """cEOSLab
Hardware version: 
Serial number: 
Hardware MAC address: 0242.0af6.c8a5
System MAC address: 0242.0af6.c8a5

Software image version: 4.26.0F-21792469.4260F (engineering build)
Architecture: i686
Internal build version: 4.26.0F-21792469.4260F
Internal build ID: c5b41f65-54cd-44b1-b576-b5c48700ee19

cEOS tools version: 1.1
Kernel version: 3.10.0-1160.21.1.el7.x86_64

Uptime: 0 weeks, 0 days, 0 hours and 8 minutes
Total memory: 4918832 kB
Free memory: 1816320 kB"""

SHOW_IP_INT_BR = """
                                                                          Address 
Interface       IP Address        Status       Protocol            MTU    Owner   
--------------- ----------------- ------------ -------------- ----------- ------- 
Ethernet1       10.0.1.4/24       up           up                 9200            
Loopback1       1.1.1.1/24        up           up                65535            
Loopback2       2.2.2.2/24        up           up                65535            
Loopback3       1.2.3.4/24        up           up                65535            

"""

commands = {
    "enable": {
        "output": None,
        "new_prompt": "{base_prompt}#",
        "help": "Enable commands for a specified privilege level",
        "prompt": INITIAL_PROMPT,
    },
    "show clock": {
        "output": make_show_clock,
        "help": "System time",
        "prompt": [INITIAL_PROMPT, "{base_prompt}#"],
    },
    "show running-config": {
        "output": RUNNING_CONFIGURATION,
        "help": "System running configuration",
        "prompt": "{base_prompt}#",
    },
    "show version": {
        "output": SHOW_VERSION,
        "help": "Software and hardware versions",
        "prompt": "{base_prompt}#",
    },
    "_default_": {
        "output": "% Invalid input",
        "help": "Output to print for unknown commands",
        "prompt": ["{base_prompt}#", "{base_prompt}>"],
    },
    "terminal width 511": {
        "output": "Width set to 511 columns.",
        "help": "Configure the terminal width to 511",
        "prompt": ["{base_prompt}#", "{base_prompt}>"],
    },
    "term width 0": {
        "alias": "terminal width 511",
    },
    "terminal length 0": {
        "output": "Pagination disabled.",
        "help": "Configure the pagination length to 0",
        "prompt": ["{base_prompt}#", "{base_prompt}>"],
    },
    "term length 0": {"alias": "terminal length 0"},
    "show ip int brief": {
        "output": SHOW_IP_INT_BR,
        "help": "Condensed output",
        "prompt": ["{base_prompt}#", "{base_prompt}>"],
    },
    "show ip interface brief": {
        "output": SHOW_IP_INT_BR,
        "help": "Condensed output",
        "prompt": ["{base_prompt}#", "{base_prompt}>"],
    },
    "conf t": {
        "prompt": "{base_prompt}#",
        "output": None,
        "new_prompt": "{base_prompt}(config)#",
        "help": "Config mode",
    },
    "config term": {"alias": "conf t"},
    "configure terminal": {"alias": "conf t"},
    "do show ip int brief": {"alias": "show ip int brief", "prompt": "{base_prompt}(config)#"},
    "exit": {"output": make_exit, "help": "Leave Exec mode", "prompt": ["{base_prompt}#", "{base_prompt}(config)#"]},
    "show hostname": {
        "output": """Hostname: {base_prompt}
FQDN:     {base_prompt}""",
        "help": "Show the system hostname",
        "prompt": "{base_prompt}#",
    },
    "no logging console": {"output": "", "help": "Set console logging parameters", "prompt": "{base_prompt}(config)#"},
    "end": {
        "output": "",
        "help": "Leave config mode",
        "new_prompt": "{base_prompt}#",
        "prompt": "{base_prompt}(config)#",
    },
}



"""
This module sets up the host object which is the main object in FakeNOS.
It provides the methods to start and stop the server instance for the host.
It also validates the host object using pydantic.
"""

import logging

from fakenos.core.pydantic_models import ModelHost
from fakenos.core.nos import available_platforms

log = logging.getLogger(__name__)


class Host:
    """
    Host class to build host instances to use with FakeNOS.
    """

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        name: str,
        username: str,
        password: str,
        port: int,
        server: dict,
        shell: dict,
        nos: dict,
        fakenos,
        platform: str = None,
    ) -> None:
        self.name: str = name
        self.server_inventory: dict = server
        self.shell_inventory: dict = shell
        self.nos_inventory: dict = nos
        self.username: str = username
        self.password: str = password
        self.port: int = port
        self.fakenos = fakenos  # FakeNOS object
        self.shell_inventory["configuration"].setdefault("base_prompt", self.name)
        self.running = False
        self.server = None
        self.server_plugin = None
        self.shell_plugin = None
        self.nos_plugin = None
        self.platform: str = platform

        if self.platform:
            self.nos_inventory["plugin"] = self.platform

        self._validate()

    def start(self):
        """Method to start server instance for this hosts"""
        self.server_plugin = self.fakenos.servers_plugins[self.server_inventory["plugin"]]
        self.shell_plugin = self.fakenos.shell_plugins[self.shell_inventory["plugin"]]
        if self.platform:
            self.nos_inventory["plugin"] = self.platform
        self.nos_plugin = self.fakenos.nos_plugins[self.nos_inventory["plugin"]]
        self.server = self.server_plugin(
            shell=self.shell_plugin,
            shell_configuration=self.shell_inventory["configuration"],
            nos=self.nos_plugin,
            nos_inventory_config=self.nos_inventory.get("configuration", {}),
            port=self.port,
            username=self.username,
            password=self.password,
            **self.server_inventory["configuration"],
        )
        self.server.start()
        self.running = True

    def stop(self):
        """Method to stop server instance of this host"""
        self.server.stop()
        self.server = None
        self.running = False

    def _validate(self):
        """Validate that the host has the required attributes using pydantic"""
        self._check_if_platform_is_supported(self.platform)
        ModelHost(**self.__dict__)

    def _check_if_platform_is_supported(self, platform: str):
        """Check if the platform is supported"""
        if platform not in available_platforms:
            raise ValueError(
                f"Platform {platform} is not supported by FakeNOS. \
                    Supported platforms are: {available_platforms}"
            )
